# init the package folder
from .mmrhist import (
    auxilary_frames,
    draw_frames,
    dynamic_timings,
    frame_position,
    get_time_offset,
    mmrhist,
    randoms,
    split_frames,
)

#from .pviews import video_frm, video_dyn



"""Simulations for image reconstruction with recommended reduced axial field of view"""
import logging

import numpy as np
from scipy import ndimage as ndi
from tqdm.auto import trange

from niftypet import nimpa

from .. import mmraux
from ..img import mmrimg
from . import mmrprj, mmrrec, petprj

__author__      = ("Pawel J. Markiewicz", "Casper O. da Costa-Luis")
__copyright__   = "Copyright 2020"
log = logging.getLogger(__name__)


def simulate_sino(
    petim,
    ctim,
    scanner_params,
    simulate_3d = False,
    slice_idx=-1,
    mu_input = False,
):
    '''
    Simulate the measured sinogram with photon attenuation.

    petim  : the input PET image based on which the emission sinogram is found
    ctim  : CT image, in register with PET and the same dimensions, is used
        for estimating the attenuation factors, which are then applied to
        simulate emission sinogram with realistic photon attenuation.
    slice_idx  : chosen 2D slice out of the 3D image for the fast simulation.
    scanner_params  : scanner parameters containing scanner constants and
        axial and transaxial look up tables (LUTs)
    mu_input  : if True, the values are representative of a mu-map in [1/cm],
        otherwise it represents the CT in [HU].
    '''
    #> decompose the scanner constants and LUTs for easier access
    Cnt = scanner_params['Cnt']

    if petim.shape != ctim.shape:
        raise ValueError('The shapes of the PET and CT images are inconsistent.')

    if simulate_3d:
        if petim.ndim != 3 \
                or petim.shape != (Cnt['SO_IMZ'], Cnt['SO_IMY'], Cnt['SO_IMX']):
            raise ValueError(
                'The input image shape does not match the scanner image size.')
        if petim.max()>200:
            log.warning('the PET image may have too large intensities for robust simulation.')
    else:
        #> 2D case with reduced rings
        if len(petim.shape) == 3:
            # make sure that the shape of the input image matches the image size of the scanner
            if petim.shape[1:] != (Cnt['SO_IMY'], Cnt['SO_IMX']):
                raise ValueError('The input image shape for x and y does not match the scanner image size.')
            # pick the right slice index (slice_idx) if not given or mistaken
            if slice_idx < 0:
                log.warning('the axial index <slice_idx> is chosen to be in the middle of axial FOV.')
                slice_idx = petim.shape[0]/2
            if slice_idx >= petim.shape[0]:
                raise ValueError('The axial index for 2D slice selection is outside the image.')
        elif len(petim.shape)==2:
            # make sure that the shape of the input image matches the image size of the scanner
            if petim.shape != (Cnt['SO_IMY'], Cnt['SO_IMX']):
                raise ValueError('The input image shape for x and y does not match the scanner image size.')
            petim.shape = (1,) + petim.shape
            ctim.shape  = (1,) + ctim.shape
            slice_idx = 0

        if 'rSZ_IMZ' not in Cnt:
            raise ValueError('Missing reduced axial FOV parameters.')

    # import pdb; pdb.set_trace()

    #--------------------
    if mu_input:
        mui = ctim
    else:
        #> get the mu-map [1/cm] from CT [HU]
        mui = nimpa.ct2mu(ctim)

    #> get rid of negative values
    mui[mui<0] = 0
    #--------------------

    if simulate_3d:
        rmu = mui
        rpet = petim
    else:
        #> 2D case with reduced rings
        #--------------------
        #> create a number of slices of the same chosen image slice for reduced (fast) 3D simulation
        rmu = mui[slice_idx,:,:]
        rmu.shape = (1,) + rmu.shape
        rmu = np.repeat(rmu, Cnt['rSZ_IMZ'], axis=0)
        #--------------------

        #--------------------
        #> form a short 3D image of the same emission image slice
        rpet = petim[slice_idx,:,:].copy()
        rpet.shape = (1,) + rpet.shape
        rpet = np.repeat(rpet, Cnt['rSZ_IMZ'], axis=0)
        #--------------------

    #> forward project the mu-map to obtain attenuation factors
    attsino = mmrprj.frwd_prj(rmu,  scanner_params, attenuation=True)

    #> forward project the PET image to obtain non-attenuated emission sino
    emisino = mmrprj.frwd_prj(rpet, scanner_params, attenuation=False)

    #> return the simulated emission sino with photon attenuation
    return attsino*emisino


def simulate_recon(
    measured_sino,
    ctim,
    scanner_params,
    simulate_3d = False,
    nitr=60,
    fwhm_rm=0.,
    slice_idx = -1,
    randoms=None,
    scatter=None,
    mu_input = False,
    msk_radius = 29.,
    psf=None,
):
    '''
    Reconstruct PET image from simulated input data
    using the EM-ML (2D) or OSEM (3D) algorithm.

    measured_sino  : simulated emission data with photon attenuation
    ctim  : either a 2D CT image or a 3D CT image from which a 2D slice
        is chosen (slice_idx) for estimation of the attenuation factors
    slice_idx  : index to extract one 2D slice for this simulation
        if input image is 3D
    nitr  : number of iterations used for the EM-ML reconstruction algorithm
    scanner_params  : scanner parameters containing scanner constants and
        axial and transaxial look up tables (LUTs)
    randoms  : randoms and scatter events (optional)
    '''
    #> decompose the scanner constants and LUTs for easier access
    Cnt = scanner_params['Cnt']
    txLUT = scanner_params['txLUT']
    axLUT = scanner_params['axLUT']
    psfkernel = mmrrec.psf_config(psf, Cnt)

    if simulate_3d:
        if ctim.ndim!=3 \
                or ctim.shape!=(Cnt['SO_IMZ'], Cnt['SO_IMY'], Cnt['SO_IMX']):
            raise ValueError(
                'The CT/mu-map image does not match the scanner image shape.')
    else:
        #> 2D case with reduced rings
        if len(ctim.shape)==3:
            # make sure that the shape of the input image matches the image size of the scanner
            if ctim.shape[1:]!=(Cnt['SO_IMY'], Cnt['SO_IMX']):
                raise ValueError('The input image shape for x and y does not match the scanner image size.')
            # pick the right slice index (slice_idx) if not given or mistaken
            if slice_idx<0:
                log.warning('the axial index <slice_idx> is chosen to be in the middle of axial FOV.')
                slice_idx = ctim.shape[0]/2
            if slice_idx>=ctim.shape[0]:
                raise ValueError('The axial index for 2D slice selection is outside the image.')
        elif len(ctim.shape)==2:
            # make sure that the shape of the input image matches the image size of the scanner
            if ctim.shape != (Cnt['SO_IMY'], Cnt['SO_IMX']):
                raise ValueError('The input image shape for x and y does not match the scanner image size.')
            ctim.shape  = (1,) + ctim.shape
            slice_idx = 0

        if 'rSZ_IMZ' not in Cnt:
            raise ValueError('Missing reduced axial FOV parameters.')

    #--------------------
    if mu_input:
        mui = ctim
    else:
        #> get the mu-map [1/cm] from CT [HU]
        mui = nimpa.ct2mu(ctim)

    #> get rid of negative values
    mui[mui<0] = 0
    #--------------------

    if simulate_3d:
        rmu = mui
        #> number of axial sinograms
        nsinos = Cnt['NSN11']
    else:
        #--------------------
        #> create a number of slides of the same chosen image slice for reduced (fast) 3D simulation
        rmu = mui[slice_idx,:,:]
        rmu.shape = (1,) + rmu.shape
        rmu = np.repeat(rmu, Cnt['rSZ_IMZ'], axis=0)
        #--------------------
        #> number of axial sinograms
        nsinos = Cnt['rNSN1']

    # import pdb; pdb.set_trace()

    #> attenuation factor sinogram
    attsino = mmrprj.frwd_prj(rmu,  scanner_params, attenuation=True, dev_out=True)

    nrmsino = np.ones(attsino.shape, dtype=np.float32)

    #> randoms and scatter put together
    if isinstance(randoms, np.ndarray) and measured_sino.shape==randoms.shape:
        rsng = mmraux.remgaps(randoms, txLUT, Cnt)
    else:
        rsng = 1e-5*np.ones((Cnt['Naw'], nsinos), dtype=np.float32)

    if isinstance(scatter, np.ndarray) and measured_sino.shape==scatter.shape:
        ssng = mmraux.remgaps(scatter, txLUT, Cnt)
    else:
        ssng = 1e-5*np.ones((Cnt['Naw'], nsinos), dtype=np.float32)

    # resolution modelling
    Cnt['SIGMA_RM'] = mmrrec.fwhm2sig(fwhm_rm, voxsize=Cnt['SZ_VOXZ']*10) if fwhm_rm else 0

    if simulate_3d:
        log.debug('------ OSEM (%d) -------' % nitr)

        # measured sinogram in GPU-enabled shape
        psng = mmraux.remgaps(measured_sino.astype(np.uint16), txLUT, Cnt)

        #> mask for reconstructed image.  anything outside it is set to zero
        msk = mmrimg.get_cylinder(Cnt, rad=msk_radius, xo=0, yo=0, unival=1, gpu_dim=True)>0.9

        #> init image
        eimg = np.ones((Cnt['SZ_IMY'], Cnt['SZ_IMX'], Cnt['SZ_IMZ']), dtype=np.float32)

        #------------------------------------
        Sn = 14 # number of subsets
        #-get one subset to get number of projection bins in a subset
        Sprj, s = mmrrec.get_subsets14(0,scanner_params)
        Nprj = len(Sprj)

        #> init subset array and sensitivity image for a given subset
        sinoTIdx = np.zeros((Sn, Nprj+1), dtype=np.int32)

        #> init sensitivity images for each subset
        sim = np.zeros((Sn, Cnt['SZ_IMY'], Cnt['SZ_IMX'], Cnt['SZ_IMZ']), dtype=np.float32)

        for n in trange(Sn, desc="sensitivity", leave=log.getEffectiveLevel() < logging.INFO):
            sinoTIdx[n,0] = Nprj #first number of projection for the given subset
            sinoTIdx[n,1:], s = mmrrec.get_subsets14(n,scanner_params)
            #> sensitivity image
            petprj.bprj(
                sim[n,:,:,:],
                attsino[sinoTIdx[n,1:],:],
                txLUT,
                axLUT,
                sinoTIdx[n,1:],
                Cnt)
        #-------------------------------------

        for k in trange(nitr, desc="OSEM",
              disable=log.getEffectiveLevel() > logging.INFO,
              leave=log.getEffectiveLevel() < logging.INFO):
            petprj.osem(
                eimg,
                psng,
                rsng,
                ssng,
                nrmsino,
                attsino,
                sinoTIdx,
                sim,
                msk,
                psfkernel,
                txLUT,
                axLUT,
                Cnt)
        eim = mmrimg.convert2e7(eimg, Cnt)

    else:
        def psf(x, output=None):
            if Cnt['SIGMA_RM']:
                x = ndi.gaussian_filter(x, sigma=Cnt['SIGMA_RM'], mode='constant', output=None)
            return x

        #> estimated image, initialised to ones
        eim = np.ones(rmu.shape, dtype=np.float32)

        msk = mmrimg.get_cylinder(Cnt, rad=msk_radius, xo=0, yo=0, unival=1, gpu_dim=False)>0.9

        #> sensitivity image for the EM-ML reconstruction
        sim = mmrprj.back_prj(attsino, scanner_params)
        sim_inv = 1 / psf(sim)
        sim_inv[~msk] = 0

        rndsct = rsng + ssng
        for i in trange(nitr, desc="MLEM",
              disable=log.getEffectiveLevel() > logging.INFO,
              leave=log.getEffectiveLevel() < logging.INFO):
            #> remove gaps from the measured sinogram
            #> then forward project the estimated image
            #> after which divide the measured sinogram by the estimated sinogram (forward projected)
            crrsino = mmraux.remgaps(measured_sino, txLUT, Cnt) / \
                        (mmrprj.frwd_prj(psf(eim), scanner_params, dev_out=True) + rndsct)

            #> back project the correction factors sinogram
            bim = mmrprj.back_prj(crrsino, scanner_params)
            bim = psf(bim, output=bim)

            #> divide the back-projected image by the sensitivity image
            #> update the estimated image and remove NaNs
            eim *= bim * sim_inv
            eim[np.isnan(eim)] = 0

    return eim



from os import getenv, path
from pathlib import Path

import pytest

HOME = Path(getenv("DATA_ROOT", "~")).expanduser()

@pytest.fixture(scope="session")
def folder_in():
    Ab_PET_mMR_test = HOME / "Ab_PET_mMR_test"
    if not Ab_PET_mMR_test.is_dir():
        pytest.skip(
            f"""Cannot find Ab_PET_mMR_test in ${{DATA_ROOT:-~}} ({HOME}).
Try running `python -m tests` to download it.
"""
        )
    return Ab_PET_mMR_test


@pytest.fixture(scope="session")
def folder_ref(folder_in):
    Ab_PET_mMR_ref = folder_in / "testing_reference" / "Ab_PET_mMR_ref"
    if not Ab_PET_mMR_ref.is_dir():
        pytest.skip(
            f"""Cannot find Ab_PET_mMR_ref in
${{DATA_ROOT:-~}}/testing_reference ({HOME}/testing_reference).
Try running `python -m tests` to download it.
"""
        )
    return Ab_PET_mMR_ref



import errno
import logging
from collections.abc import Iterable
from os import fspath, path
from textwrap import dedent

import numpy as np
import pytest
from filelock import FileLock

from niftypet import nimpa, nipet

# segmentation/parcellation for PVC, with unique regions numbered from 0 onwards
pvcroi = []
pvcroi.append([66, 67] + list(range(81, 95)))  # white matter
pvcroi.append([36])  # brain stem
pvcroi.append([35])  # pons
pvcroi.append([39, 40, 72, 73, 74])  # cerebellum GM
pvcroi.append([41, 42])  # cerebellum WM
pvcroi.append([48, 49])  # hippocampus
pvcroi.append([167, 168])  # posterior cingulate gyrus
pvcroi.append([139, 140])  # middle cingulate gyrus
pvcroi.append([101, 102])  # anterior cingulate gyrus
pvcroi.append([169, 170])  # precuneus
pvcroi.append([32, 33])  # amygdala
pvcroi.append([37, 38])  # caudate
pvcroi.append([56, 57])  # pallidum
pvcroi.append([58, 59])  # putamen
pvcroi.append([60, 61])  # thalamus
pvcroi.append([175, 176, 199, 200])  # parietal without precuneus
pvcroi.append([133, 134, 155, 156, 201, 202, 203, 204])  # temporal
pvcroi.append([4, 5, 12, 16, 43, 44, 47, 50, 51, 52, 53])  # CSF
pvcroi.append([24, 31, 62, 63, 70, 76, 77, 96, 97])  # basal ganglia + optic chiasm
pvcroi.append(
    list(range(103, 110 + 1))
    + list(range(113, 126 + 1))
    + list(range(129, 130 + 1))
    + list(range(135, 138 + 1))
    + list(range(141, 154 + 1))
    + list(range(157, 158 + 1))
    + list(range(161, 166 + 1))
    + list(range(171, 174 + 1))
    + list(range(177, 188 + 1))
    + list(range(191, 198 + 1))
    + list(range(205, 208 + 1))
)  # remaining neocortex
# expected %error for static (SUVr) and PVC reconstructions
emape_basic = 0.1
emape_algnd = {
    "pet": 3.0,
    "pos": 0.1,
    "trm": 3.0,
    "pvc": 3.0,
    "hmu": 0.01,
    "omu": 3.0,
}


@pytest.fixture(scope="session")
def mMRpars():
    params = nipet.get_mmrparams()
    # params["Cnt"]["VERBOSE"] = True
    params["Cnt"]["LOG"] = logging.INFO
    return params


@pytest.fixture(scope="session")
def datain(mMRpars, folder_in):
    return nipet.classify_input(folder_in, mMRpars)


@pytest.fixture(scope="session")
def muhdct(mMRpars, datain, tmp_path_factory, worker_id):
    tmp_path = tmp_path_factory.getbasetemp()

    if worker_id == "master":  # not xdist, auto-reuse
        opth = str(tmp_path / "muhdct")
        return nipet.hdw_mumap(
            datain, [1, 2, 4], mMRpars, outpath=opth, use_stored=True
        )

    opth = str(tmp_path.parent / "muhdct")
    flock = FileLock(opth + ".lock")
    with flock.acquire(poll_intervall=0.5):  # xdist, force auto-reuse via flock
        return nipet.hdw_mumap(
            datain, [1, 2, 4], mMRpars, outpath=opth, use_stored=True
        )


@pytest.fixture(scope="session")
def refimg(folder_ref):
    # predetermined structure of the reference folder
    basic = folder_ref / "basic"
    spm = folder_ref / "dyn_aligned" / "spm"
    niftyreg = folder_ref / "dyn_aligned" / "niftyreg"
    refpaths = {
        "histo": {"p": 1570707830, "d": 817785422},
        "basic": {
            "pet": basic / "17598013_t-3000-3600sec_itr-4_suvr.nii.gz",
            "omu": basic / "mumap-from-DICOM_no-alignment.nii.gz",
            "hmu": basic / "hardware_umap.nii.gz",
        },
        "aligned": {
            "spm": {
                "hmu": spm / "hardware_umap.nii.gz",
                "omu": spm / "mumap-PCT-aligned-to_t0-3600_AC.nii.gz",
                "pos": spm / "17598013_t0-3600sec_itr2_AC-UTE.nii.gz",
                "pet": spm / "17598013_nfrm-2_itr-4.nii.gz",
                "trm": spm / "17598013_nfrm-2_itr-4_trimmed-upsampled-scale-2.nii.gz",
                "pvc": spm / "17598013_nfrm-2_itr-4_trimmed-upsampled-scale-2_PVC.nii.gz",
            },
            "niftyreg": {
                "hmu": niftyreg / "hardware_umap.nii.gz",
                "omu": niftyreg / "mumap-PCT-aligned-to_t0-3600_AC.nii.gz",
                "pos": niftyreg / "17598013_t0-3600sec_itr2_AC-UTE.nii.gz",
                "pet": niftyreg / "17598013_nfrm-2_itr-4.nii.gz",
                "trm": niftyreg / "17598013_nfrm-2_itr-4_trimmed-upsampled-scale-2.nii.gz",
                "pvc": niftyreg / "17598013_nfrm-2_itr-4_trimmed-upsampled-scale-2_PVC.nii.gz",
            },
        },
    }

    testext = {
        "basic": {
            "pet": "static reconstruction with unaligned UTE mu-map",
            "hmu": "hardware mu-map for the static unaligned reconstruction",
            "omu": "object mu-map for the static unaligned reconstruction",
        },
        "aligned": {
            "hmu": "hardware mu-map for the 2-frame aligned reconstruction",
            "omu": "object mu-map for the 2-frame aligned reconstruction",
            "pos": "AC reconstruction for positioning (full acquisition used)",
            "pet": "2-frame scan with aligned UTE mu-map",
            "trm": "trimming post reconstruction",
            "pvc": "PVC post reconstruction",
        },
    }

    # check basic files
    frefs = refpaths["basic"]
    for k, v in frefs.items():
        if not v.is_file():
            raise FileNotFoundError(errno.ENOENT, v)

    # check reg tools: niftyreg and spm
    frefs = refpaths["aligned"]
    for r in frefs:
        for k, v in frefs[r].items():
            if not v.is_file():
                raise FileNotFoundError(errno.ENOENT, v)

    return refpaths, testext


def test_histogramming(mMRpars, datain, refimg, tmp_path):
    refpaths, _ = refimg
    opth = str(tmp_path / "histogramming")
    hst = nipet.mmrhist(datain, mMRpars, outpath=opth, store=True)

    # prompt counts: head curve & sinogram
    assert np.sum(hst["phc"]) == np.sum(hst["psino"])
    # delayed counts: head curve & sinogram
    assert np.sum(hst["dhc"]) == np.sum(hst["dsino"])
    # prompt counts: head curve & reference
    assert np.sum(hst["phc"]) == refpaths["histo"]["p"]
    # delayed counts: head curve & reference
    assert np.sum(hst["dhc"]) == refpaths["histo"]["d"]


def test_basic_reconstruction(mMRpars, datain, muhdct, refimg, tmp_path):
    refpaths, testext = refimg
    opth = str(tmp_path / "basic_reconstruction")

    muodct = nipet.obj_mumap(datain, mMRpars, outpath=opth, store=True)
    recon = nipet.mmrchain(
        datain,
        mMRpars,
        frames=["timings", [3000, 3600]],
        mu_h=muhdct,
        mu_o=muodct,
        itr=4,
        fwhm=0.0,
        outpath=opth,
        fcomment="_suvr",
        store_img=True,
        store_img_intrmd=True,
    )

    testout = {"pet": recon["fpet"], "hmu": muhdct["im"], "omu": muodct["im"]}
    for k in testext["basic"]:
        diff = nimpa.imdiff(fspath(refpaths["basic"][k]), testout[k], verbose=True, plot=False)
        assert diff["mape"] <= emape_basic, testext["basic"][k]


@pytest.mark.parametrize("reg_tool", ["niftyreg", "spm"])
def test_aligned_reconstruction(reg_tool, mMRpars, datain, muhdct, refimg, tmp_path):
    refpaths, testext = refimg
    opth = str(tmp_path / "aligned_reconstruction")

    muopct = nipet.align_mumap(
        datain,
        mMRpars,
        outpath=opth,
        t0=0,
        t1=3600,
        reg_tool=reg_tool,
        store=True,
        itr=2,
        petopt="ac",
        fcomment="",
        musrc="pct",
    )
    # matshow(muopct['im'][60,:,:] + muhdct['im'][60,:,:])
    recon = nipet.mmrchain(
        datain,
        mMRpars,
        frames=["timings", [80, 260], [3000, 3600]],
        mu_h=muhdct,
        mu_o=muopct,
        itr=4,
        fwhm=0.0,
        outpath=opth,
        trim=True,
        pvcroi=pvcroi,
        pvcreg_tool=reg_tool,
        store_img=True,
        store_img_intrmd=True,
    )

    testout = {
        "pet": recon["fpet"],
        "hmu": muhdct["im"],
        "omu": muopct["im"],
        "pos": muopct["fpet"],
        "trm": recon["trimmed"]["fpet"],
        "pvc": recon["trimmed"]["fpvc"],
    }
    for k in testext["aligned"]:
        diff = nimpa.imdiff(
            fspath(refpaths["aligned"][reg_tool][k]), testout[k], verbose=True, plot=False
        )
        err = diff["mape"] <= emape_algnd[k]
        assert (all(err) if isinstance(err, Iterable) else err), testext["aligned"][k]



#!/usr/bin/env python
"""initialise the NiftyPET NIPET package"""
__author__      = ("Pawel J. Markiewicz", "Casper O. da Costa-Luis")
__copyright__   = "Copyright 2020"
# version detector. Precedence: installed dist, git, 'UNKNOWN'
try:
    from ._dist_ver import __version__
except ImportError:
    try:
        from setuptools_scm import get_version

        __version__ = get_version(root="../..", relative_to=__file__)
    except (ImportError, LookupError):
        __version__ = "UNKNOWN"

import logging
import os
import platform
import re
import sys
from textwrap import dedent

from pkg_resources import resource_filename
from tqdm.auto import tqdm

from niftypet.ninst import cudasetup as cs
from niftypet.ninst.dinf import dev_info, gpuinfo
from niftypet.ninst.tools import LOG_FORMAT, LogHandler, path_resources, resources

# shared CUDA C library for extended auxiliary functions for the mMR
#> Siemens Biograph mMR
from . import img, lm, mmr_auxe, mmraux, mmrnorm, prj
from .img.mmrimg import align_mumap
from .img.mmrimg import convert2dev as im_e72dev
from .img.mmrimg import convert2e7 as im_dev2e7
from .img.mmrimg import hdw_mumap, obj_mumap, pct_mumap
from .img.pipe import mmrchain
from .lm.mmrhist import dynamic_timings, mmrhist, randoms
from .mmraux import explore_input as classify_input
from .mmraux import mMR_params as get_mmrparams
from .prj.mmrprj import back_prj, frwd_prj
from .prj.mmrsim import simulate_recon, simulate_sino
from .sct.mmrsct import vsm

# log = logging.getLogger(__name__)
# technically bad practice to add handlers
# https://docs.python.org/3/howto/logging.html#library-config
# log.addHandler(LogHandler())  # do it anyway for convenience










if resources.ENBLAGG:
    from .lm.pviews import video_dyn, video_frm

if resources.ENBLXNAT:
    from xnat import xnat


#> GE Signa
#from . import aux_sig

#from . import lm_sig
#from .lm_sig.hst_sig import lminfo_sig

# for use in `cmake -DCMAKE_PREFIX_PATH=...`
cmake_prefix = resource_filename(__name__, "cmake")



"""Forward and back projector for PET data reconstruction"""
import logging
import os
import sys

import numpy as np

from .. import mmraux
from ..img import mmrimg
from . import petprj

__author__      = ("Pawel J. Markiewicz", "Casper O. da Costa-Luis")
__copyright__   = "Copyright 2020"
log = logging.getLogger(__name__)


#=========================================================================
# transaxial (one-slice) projector
#-------------------------------------------------------------------------


def trnx_prj(scanner_params, sino=None, im=None):

    # Get particular scanner parameters: Constants, transaxial and axial LUTs
    Cnt   = scanner_params['Cnt']
    txLUT = scanner_params['txLUT']
    axLUT = scanner_params['axLUT']

    # if sino==None and im==None:
    #     raise ValueError('Input sinogram or image has to be given.')
    if sino is not None and im is not None:
        raise ValueError('Only one input should be given: sinogram or image.')

    if sino is None:
        sino = np.zeros((txLUT['Naw'], ), dtype=np.float32)
    if im is None:
        im = np.zeros((Cnt['SO_IMY'], Cnt['SO_IMX']), dtype=np.float32)

    tv = np.zeros(Cnt['NTV']*Cnt['Naw'], dtype=np.uint8)
    tt = np.zeros(Cnt['NTT']*Cnt['Naw'], dtype=np.float32)

    petprj.tprj(sino, im, tv, tt, txLUT, Cnt)

    return {'tv':tv, 'tt':tt}


#=========================================================================
# forward projector
#-------------------------------------------------------------------------


def frwd_prj(im, scanner_params, isub=np.array([-1], dtype=np.int32), dev_out=False, attenuation=False):
    ''' Calculate forward projection (a set of sinograms) for the provided input image.
        Arguments:
        im -- input image (can be emission or mu-map image).
        scanner_params -- dictionary of all scanner parameters, containing scanner constants,
            transaxial and axial look up tables (LUT).
        isub -- array of transaxial indices of all sinograms (angles x bins) used for subsets.
            when the first element is negative, all transaxial bins are used (as in pure EM-ML).
        dev_out -- if True, output sinogram is in the device form, i.e., with two dimensions
            (# bins/angles, # sinograms) instead of default three (# sinograms, # bins, # angles).
        attenuation -- controls whether emission or LOR attenuation probability sinogram
            is calculated; the default is False, meaning emission sinogram; for attenuation
            calculations (attenuation=True), the exponential of the negative of the integrated
            mu-values along LOR path is taken at the end.
    '''
    # Get particular scanner parameters: Constants, transaxial and axial LUTs
    Cnt   = scanner_params['Cnt']
    txLUT = scanner_params['txLUT']
    axLUT = scanner_params['axLUT']

    #>choose between attenuation forward projection (mu-map is the input)
    #>or the default for emission image forward projection
    if attenuation:
        att = 1
    else:
        att = 0

    if Cnt['SPN']==1:
        # number of rings calculated for the given ring range (optionally we can use only part of the axial FOV)
        NRNG_c = Cnt['RNG_END'] - Cnt['RNG_STRT']
        # number of sinos in span-1
        nsinos = NRNG_c**2
        # correct for the max. ring difference in the full axial extent (don't use ring range (1,63) as for this case no correction)
        if NRNG_c==64:
            nsinos -= 12
    elif  Cnt['SPN']==11: nsinos=Cnt['NSN11']
    elif  Cnt['SPN']==0:  nsinos=Cnt['NSEG0']

    if im.shape[0]==Cnt['SO_IMZ'] and im.shape[1]==Cnt['SO_IMY'] and im.shape[2]==Cnt['SO_IMX']:
        ims = mmrimg.convert2dev(im, Cnt)
    elif im.shape[0]==Cnt['SZ_IMX'] and im.shape[1]==Cnt['SZ_IMY'] and im.shape[2]==Cnt['SZ_IMZ']:
        ims = im
    elif im.shape[0]==Cnt['rSO_IMZ'] and im.shape[1]==Cnt['SO_IMY'] and im.shape[2]==Cnt['SO_IMX']:
        ims = mmrimg.convert2dev(im, Cnt)
    elif im.shape[0]==Cnt['SZ_IMX'] and im.shape[1]==Cnt['SZ_IMY'] and im.shape[2]==Cnt['rSZ_IMZ']:
        ims = im
    else:
        raise ValueError('wrong image size;'
            ' it has to be one of these: (z,y,x) = (127,344,344)'
            ' or (y,x,z) = (320,320,128)')

    log.debug('number of sinos:%d' % nsinos)

    #predefine the sinogram.  if subsets are used then only preallocate those bins which will be used.
    if isub[0]<0:
        sinog = np.zeros((txLUT['Naw'], nsinos), dtype=np.float32)
    else:
        sinog = np.zeros((len(isub), nsinos), dtype=np.float32)

    # --------------------
    petprj.fprj(sinog, ims, txLUT, axLUT, isub, Cnt, att)
    # --------------------
    # get the sinogram bins in a proper sinogram
    sino = np.zeros((txLUT['Naw'], nsinos), dtype=np.float32)
    if isub[0]>=0:    sino[isub,:] = sinog
    else:  sino = sinog

    # put the gaps back to form displayable sinogram
    if not dev_out:
        sino = mmraux.putgaps(sino, txLUT, Cnt)

    return sino


#=========================================================================
# back projector
#-------------------------------------------------------------------------


def back_prj(sino, scanner_params, isub=np.array([-1], dtype=np.int32)):
    '''
    Calculate forward projection for the provided input image.
    Arguments:
        sino -- input emission sinogram to be back projected to the image space.
        scanner_params -- dictionary of all scanner parameters, containing scanner constants,
            transaxial and axial look up tables (LUT).
        isub -- array of transaxial indices of all sinograms (angles x bins) used for subsets;
            when the first element is negative, all transaxial bins are used (as in pure EM-ML).
    '''
    # Get particular scanner parameters: Constants, transaxial and axial LUTs
    Cnt   = scanner_params['Cnt']
    txLUT = scanner_params['txLUT']
    axLUT = scanner_params['axLUT']

    if Cnt['SPN']==1:
        # number of rings calculated for the given ring range (optionally we can use only part of the axial FOV)
        NRNG_c = Cnt['RNG_END'] - Cnt['RNG_STRT']
        # number of sinos in span-1
        nsinos = NRNG_c**2
        # correct for the max. ring difference in the full axial extent (don't use ring range (1,63) as for this case no correction)
        if NRNG_c==64:
            nsinos -= 12
    elif  Cnt['SPN']==11: nsinos=Cnt['NSN11']
    elif  Cnt['SPN']==0:  nsinos=Cnt['NSEG0']


    #> check first the Siemens default sinogram;
    #> for this default shape only full sinograms are expected--no subsets.
    if len(sino.shape)==3:
        if sino.shape[0]!=nsinos or sino.shape[1]!=Cnt['NSANGLES'] or sino.shape[2]!=Cnt['NSBINS']:
            raise ValueError('Unexpected sinogram array dimensions/shape for Siemens defaults.')
        sinog = mmraux.remgaps(sino, txLUT, Cnt)

    elif len(sino.shape)==2:
        if isub[0]<0 and sino.shape[0]!=txLUT["Naw"]:
            raise ValueError('Unexpected number of transaxial elements in the full sinogram.')
        elif isub[0]>=0 and sino.shape[0]!=len(isub):
            raise ValueError('Unexpected number of transaxial elements in the subset sinogram.')
        #> check if the number of sinograms is correct
        if sino.shape[1]!=nsinos:
            raise ValueError('Inconsistent number of sinograms in the array.')
        #> when found the dimensions/shape are fine:
        sinog = sino
    else:
        raise ValueError('Unexpected shape of the input sinogram.')

    #predefine the output image depending on the number of rings used
    if Cnt['SPN']==1 and 'rSZ_IMZ' in Cnt:
        nvz = Cnt['rSZ_IMZ']
    else:
        nvz = Cnt['SZ_IMZ']
    bimg = np.zeros((Cnt['SZ_IMX'], Cnt['SZ_IMY'], nvz), dtype=np.float32)

    #> run back-projection
    petprj.bprj(bimg, sinog, txLUT, axLUT, isub, Cnt)

    #> change from GPU optimised image dimensions to the standard Siemens shape
    bimg = mmrimg.convert2e7(bimg, Cnt)

    return bimg



'''
Voxel-driven scatter modelling for PET data
'''
import logging
import os
import sys
import time
from concurrent.futures import ThreadPoolExecutor
from math import pi

import nibabel as nib
import numpy as np
import scipy.ndimage as ndi
from scipy.interpolate import CloughTocher2DInterpolator, interp2d
from scipy.spatial import qhull
from scipy.special import erfc

from .. import mmr_auxe, mmraux, mmrnorm
from ..img import mmrimg
from ..prj import mmrprj, mmrrec, petprj
from . import nifty_scatter

__author__      = ("Pawel J. Markiewicz", "Casper O. da Costa-Luis")
__copyright__   = "Copyright 2020"
log = logging.getLogger(__name__)


def fwhm2sig (fwhm, Cnt):
    '''
    Convert FWHM to sigma (standard deviation)
    '''
    return (fwhm/Cnt['SO_VXY']) / (2*(2*np.log(2))**.5)


#=======================================================================
# S C A T T E R
#-----------------------------------------------------------------------


def get_scrystals(scanner_params):
    '''
    Get table of selected transaxial and axial (ring) crystals
    used for scatter modelling
    '''
    #> decompose constants, transaxial and axial LUTs are extracted
    Cnt   = scanner_params['Cnt']
    txLUT = scanner_params['txLUT']
    axLUT = scanner_params['axLUT']

    #------------------------------------------------------
    #> transaxial crystals definitions
    crs = txLUT['crs']

    #> period of scatter crystals (needed for definition)
    SCRS_T = 7

    #> counter for crystal period, SCRS_T
    cntr = 0

    #> scatter crystal index
    iscrs = 0

    #> initialise list of transaxial scatter crystal table
    scrs = []

    #> transaxial scatter crystal selection for modelling
    for c in range(Cnt['NCRS']):
        if (((c + 1) % 9) == 0):
            continue
        cntr += 1
        if (cntr == SCRS_T):
            cntr = 0
            scrs.append([
                c, 0.5*(crs[c, 0] + crs[c, 2]), 0.5*(crs[c, 1] + crs[c, 3])
            ])
            iscrs += 1

    #> convert the scatter crystal table to Numpy array
    scrs = np.array(scrs, dtype=np.float32)
    #------------------------------------------------------


    #------------------------------------------------------
    #> scatter ring definition (axially)
    sct_irng = np.int16([0, 10, 19, 28, 35, 44, 53, 63])
    # number of scatter rings (used for scatter estimation)
    NSRNG = len(sct_irng)
    #------------------------------------------------------


    logtxt = ''

    srng = np.zeros((NSRNG,2), dtype=np.float32)
    z = 0.5*(-Cnt['NRNG']*Cnt['AXR'] + Cnt['AXR'])
    for ir in range(NSRNG):
        srng[ir,0] = float(sct_irng[ir])
        srng[ir,1] = axLUT['rng'][sct_irng[ir],:].mean()
        logtxt += '> [{}]: ring_i={}, ring_z={}\n'.format(ir, int(srng[ir,0]), srng[ir,1])

    log.debug(logtxt)


    return dict(scrs=scrs, srng=srng, sirng=sct_irng, NSCRS=scrs.shape[0], NSRNG=NSRNG)


#=======================================================================
def get_sctlut2d(txLUT, scrs_def):

    #> scatter to sinogram bin index LUT
    sct2aw = np.zeros(scrs_def['NSCRS']*scrs_def['NSCRS'], dtype=np.int32)

    # scatter/unscattered crystal x-coordinate (used for determining +/- sino segments)
    xsxu = np.zeros((scrs_def['NSCRS'], scrs_def['NSCRS']), dtype=np.int8)

    #> loop over unscattered crystals
    for uc in range(scrs_def['NSCRS']):

        #> loop over scatter crystals
        for sc in range(scrs_def['NSCRS']):

            #> sino linear index (full including any gaps)
            #> scrs_def['scrs'] is a 2D array of rows [sct_crs_idx, mid_x, mid_y]
            sct2aw[scrs_def['NSCRS']*uc + sc] = \
                txLUT['c2sFw'][
                    int(scrs_def['scrs'][uc,0]),
                    int(scrs_def['scrs'][sc,0])
                ]

            #> scattered and unscattered crystal positions (used for determining +/- sino segments)
            xs = scrs_def['scrs'][sc,1]
            xu = scrs_def['scrs'][uc,1]

            if (xs>xu):
                xsxu[uc, sc] = 1

    sct2aw.shape = (scrs_def['NSCRS'], scrs_def['NSCRS'])

    return dict(sct2aw=sct2aw, xsxu=xsxu, c2sFw=txLUT['c2sFw'])

#=======================================================================


#=======================================================================
def get_knlut(Cnt):
    '''
    get Klein-Nishina LUTs
    '''

    SIG511 = Cnt['ER']*Cnt['E511']/2.35482

    CRSSavg = (2*(4/3.0-np.log(3)) + .5*np.log(3)-4/9.0)

    COSSTP = (1-Cnt['COSUPSMX'])/(Cnt['NCOS']-1)

    log.debug('using these scatter constants:\nCOS(UPSMAX) = {},\nCOSSTP = {}'.format(Cnt['COSUPSMX'], COSSTP))

    knlut = np.zeros((Cnt['NCOS'],2), dtype = np.float32)

    for i in range(Cnt['NCOS']):
        cosups = Cnt['COSUPSMX']+i*COSSTP
        alpha = 1/(2 - cosups)
        KNtmp = ( (0.5*Cnt['R02']) * alpha*alpha * ( alpha + 1/alpha - (1-cosups*cosups) ) )
        knlut[i,0] = KNtmp / ( 2*pi*Cnt['R02'] * CRSSavg);
        knlut[i,1] = ( (1+alpha)/(alpha*alpha)*(2*(1+alpha)/(1+2*alpha)-1/alpha*np.log(1+2*alpha)) + \
                        np.log(1+2*alpha)/(2*alpha)-(1+3*alpha)/((1+2*alpha)*(1+2*alpha)) ) / CRSSavg

        # Add energy resolution:
        if Cnt['ER']>0:
            log.info('using energy resolution for scatter simulation, ER = {}'.format(Cnt['ER']))
            knlut[i,0] *= .5*erfc( (Cnt['LLD']-alpha*Cnt['E511'])/(SIG511*np.sqrt(2*alpha)) )
            #knlut[i,0] *= .5*erfc( (Cnt['LLD']-alpha*Cnt['E511'])/(SIG511) );

        # for large angles (small cosups) when the angle in GPU calculations is greater than COSUPSMX
        if (i==0):
            knlut[0,0] = 0;

    return knlut
#=======================================================================



#==================================================================================================
# GET SCATTER LUTs
#--------------------------------------------------------------------------------------------------
def rd2sni(offseg, r1, r0):
    rd = np.abs(r1-r0)
    rdi = (2*rd - 1*(r1>r0))
    sni = offseg[rdi] + np.minimum(r0,r1)
    return sni
#--------------------------------------------------------------------------------------------------

def get_sctLUT(scanner_params):

    #> decompose constants, transaxial and axial LUTs are extracted
    Cnt   = scanner_params['Cnt']
    txLUT = scanner_params['txLUT']
    axLUT = scanner_params['axLUT']


    #> get the Klein-Nishina LUT:
    KN = get_knlut(Cnt)

    #> get scatter crystal tables:
    scrs_def = get_scrystals(scanner_params)

    #> get 2D scatter LUT (for transaxial sinograms)
    sctlut2d = get_sctlut2d(txLUT, scrs_def)

    # get the indexes of rings used for scatter estimation
    irng = scrs_def['sirng']

    # get number of ring accounting for the possible ring reduction (to save computation time)
    # NRNG = Cnt['RNG_END']-Cnt['RNG_STRT']

    #-span-1 LUT (rings to sino index)
    seg = np.append( [Cnt['NRNG']], np.ceil( np.arange(Cnt['NRNG']-1,0,-.5) ).astype(np.int16) )
    offseg = np.int16( np.append( [0], np.cumsum(seg)) )

    #-3D scatter sino LUT. axial component based on michelogram.
    sctaxR = np.zeros((Cnt['NRNG']**2, 4), dtype=np.int32)
    sctaxW = np.zeros((Cnt['NRNG']**2, 4), dtype=np.float32)

    #-just for local check and display of the interpolation at work
    mich  = np.zeros((Cnt['NRNG'], Cnt['NRNG']), dtype=np.float32)
    mich2 = np.zeros((Cnt['NRNG'], Cnt['NRNG']), dtype=np.float32)


    J, I =  np.meshgrid(irng, irng)
    mich[J,I] = np.reshape(np.arange(scrs_def['NSRNG']**2), (scrs_def['NSRNG'], scrs_def['NSRNG']))
    # plt.figure(64), plt.imshow(mich, interpolation='none')

    for r1 in range(Cnt['RNG_STRT'], Cnt['RNG_END']):
        #border up and down
        bd = next(idx for idx in irng        if idx>=r1)
        bu = next(idx for idx in irng[::-1]  if idx<=r1)
        for r0 in range(Cnt['RNG_STRT'], Cnt['RNG_END']):

            # if (np.abs(r1-r0)>MRD):
            #     continue
            #border left and right
            br = next(idx for idx in irng        if idx>=r0)
            bl = next(idx for idx in irng[::-1]  if idx<=r0)
            #print '(r0,r1)=', r0,r1, '(bl,br,bu,bd)', bl,br,bu,bd

            #span-1 sino index (sni) creation:
            sni = rd2sni(offseg, r1, r0)

            #see: https://en.wikipedia.org/wiki/Bilinear_interpolation
            if (br==bl)and(bu!=bd):

                sctaxR[sni,0] = rd2sni(offseg, bd, r0)
                sctaxW[sni,0] = (r1-bu)/float(bd-bu)
                sctaxR[sni,1] = rd2sni(offseg, bu, r0)
                sctaxW[sni,1] = (bd-r1)/float(bd-bu)

                mich2[r1,r0] = mich[bd,r0]*sctaxW[sni,0]  +  mich[bu,r0]*sctaxW[sni,1]

            elif (bu==bd)and(br!=bl):

                sctaxR[sni,0] = rd2sni(offseg, r1, bl)
                sctaxW[sni,0] = (br-r0)/float(br-bl)
                sctaxR[sni,1] = rd2sni(offseg, r1, br)
                sctaxW[sni,1] = (r0-bl)/float(br-bl)

                mich2[r1,r0] =  mich[r1,bl]*sctaxW[sni,0] + mich[r1,br]*sctaxW[sni,1]

            elif (bu==bd)and(br==bl):

                mich2[r1,r0] = mich[r1,r0]
                sctaxR[sni,0] = rd2sni(offseg, r1, r0)
                sctaxW[sni,0] = 1
                continue

            else:

                cf = float(((br-bl)*(bd-bu)))

                sctaxR[sni,0] = rd2sni(offseg, bd, bl)
                sctaxW[sni,0] = (br-r0)*(r1-bu)/cf
                sctaxR[sni,1] = rd2sni(offseg, bd, br)
                sctaxW[sni,1] = (r0-bl)*(r1-bu)/cf

                sctaxR[sni,2] = rd2sni(offseg, bu, bl)
                sctaxW[sni,2] = (br-r0)*(bd-r1)/cf
                sctaxR[sni,3] = rd2sni(offseg, bu, br)
                sctaxW[sni,3] = (r0-bl)*(bd-r1)/cf

                mich2[r1,r0] =  mich[bd,bl]*sctaxW[sni,0]+ mich[bd,br]*sctaxW[sni,1] + mich[bu,bl]*sctaxW[sni,2] + mich[bu,br]*sctaxW[sni,3]

    # plt.figure(65), plt.imshow(mich2, interpolation='none')


    sctLUT = {
        'sctaxR':sctaxR,
        'sctaxW':sctaxW,
        'offseg':offseg,
        'KN':KN,
        'mich_chck':[mich, mich2],
        **scrs_def,
        **sctlut2d,
        }

    return sctLUT


#-------------------------------------------------------------------------------------------------
# S C A T T E R    I N T E R P O L A T I O N
#-------------------------------------------------------------------------------------------------


#==============================================================================
def intrp_bsct(sct3d, Cnt, sctLUT, ssrlut, dtype=np.float32):
    '''
    interpolate the basic scatter distributions which are then
    transferred into the scatter sinograms.
    '''


    #> number of sinograms
    snno = sct3d.shape[1]

    i_scrs = sctLUT['scrs'][:,0].astype(int)

    x = i_scrs
    y = np.append([-1], i_scrs)
    xnew = np.arange(Cnt['NCRS'])
    ynew = np.arange(Cnt['NCRS'])

    #> advanced indexing matrix for rolling the non-interpolated results
    jj, ii = np.mgrid[0:sctLUT['NSCRS'], 0:sctLUT['NSCRS']]

    #> roll each row according to the position
    for i in range(sctLUT['NSCRS']):
        ii[i,:] = np.roll(ii[i,:], -1*i)

    jjnew, iinew = np.mgrid[0:Cnt['NCRS'], 0:Cnt['NCRS']]
    for i in range(Cnt['NCRS']):
        iinew[i,:] = np.roll(iinew[i,:], i)

    ssn = np.zeros((Cnt['TOFBINN'], snno, Cnt['NSANGLES'], Cnt['NSBINS']), dtype=dtype);
    sssr = np.zeros((Cnt['TOFBINN'], Cnt['NSEG0'], Cnt['NSANGLES'], Cnt['NSBINS']), dtype=dtype);


    for ti in range(Cnt['TOFBINN']):
        sn2d = np.zeros(Cnt['NSANGLES']*Cnt['NSBINS'], dtype=dtype)

        for si in range(snno):

            sn2d[:] = 0

            sct2d = sct3d[0, si, jj, ii]

            z = np.vstack([sct2d[-1,:], sct2d])
            f = interp2d(x, y, z, kind='cubic')
            znew = f(xnew, ynew)

            # unroll
            znew = znew[jjnew, iinew]

            #> upper triangle
            #> add '1' to include index zero (distinguished from after triangulation)
            qi = np.triu(sctLUT['c2sFw']+1)>0
            sidx = sctLUT['c2sFw'][qi]
            s = znew[qi]
            sn2d[sidx] = s

            #> lower triangle
            qi = np.tril(sctLUT['c2sFw']+1)>0
            sidx = sctLUT['c2sFw'][qi]
            s = znew[qi]
            sn2d[sidx] += s

            ssn [ti, si, ...] = np.reshape(sn2d, (Cnt['NSANGLES'],Cnt['NSBINS']))
            sssr[ti, ssrlut[si], ...] += ssn[ti, si,:,:]


    return np.squeeze(ssn), np.squeeze(sssr)
    #-------------------------------------------------

#====================================================================================================


def vsm(
        datain,
        mumaps,
        em,
        scanner_params,
        histo = None,
        rsino = None,
        prcnt_scl = 0.1,
        fwhm_input=0.42,
        mask_threshlod = 0.999,
        snmsk=None,
        emmsk=False,
        interpolate=True,
        return_uninterp=False,
        return_ssrb=False,
        return_mask=False,
        return_scaling=False,
        scaling=True,
        self_scaling=False,
        save_sax=False,
    ):
    '''
    Voxel-driven scatter modelling (VSM).
    Obtain a scatter sinogram using the mu-maps (hardware and object mu-maps)
    an estimate of emission image, the prompt measured sinogram, an
    estimate of the randoms sinogram and a normalisation sinogram.
    Input:
        - datain:       Contains the data used for scatter-specific detector
                        normalisation.  May also include the non-corrected
                        emission image used for masking, when requested.
        - mumaps:       A tuple of hardware and object mu-maps (in this order).
        - em:           An estimate of the emission image.
        - histo:          Dictionary containing the histogrammed measured data into
                        sinograms.
        - rsino:       Randoms sinogram (3D).  Needed for proper scaling of
                        scatter to the prompt data.
        - scanner_params: Scanner specific parameters.
        - prcnt_scl:    Ratio of the maximum scatter intensities below which the
                        scatter is not used for fitting it to the tails of prompt
                        data.  Default is 10%.
        - emmsk:        When 'True' it will use uncorrected emission image for
                        masking the sources (voxels) of photons to be used in the
                        scatter modelling.
        - scaling:      performs scaling to the data (sinogram)
        - self_scaling: Scaling is performed on span-1 without the help of SSR
                        scaling and using the sax factors (scatter axial factors).
                        If False (default), the sax factors have to be provided.
        - sax:          Scatter axial factors used for scaling with SSR sinograms.

    '''

    #> decompose constants, transaxial and axial LUTs are extracted
    Cnt   = scanner_params['Cnt']
    txLUT = scanner_params['txLUT']
    axLUT = scanner_params['axLUT']

    if self_scaling:
        scaling = True

    #> decompose mu-maps
    muh, muo = mumaps

    if emmsk and not os.path.isfile(datain['em_nocrr']):
        log.info('reconstructing emission data without scatter and attenuation corrections for mask generation...')
        recnac = mmrrec.osemone(datain, mumaps, histo, scanner_params, recmod=0, itr=3, fwhm=2.0, store_img=True)
        datain['em_nocrr'] = recnac.fpet

    # if rsino is None and not histo is None and 'rsino' in histo:
    #     rsino = histo['rsino']

    #> if histogram data or randoms sinogram not given, then no scaling or normalisation
    if (histo is None) or (rsino is None):
        scaling = False

    #-get the normalisation components
    nrmcmp, nhdr = mmrnorm.get_components(datain, Cnt)

    #-smooth for defining the sino scatter only regions
    if fwhm_input>0.:
        mu_sctonly =  ndi.filters.gaussian_filter(
            mmrimg.convert2dev(muo, Cnt),
            fwhm2sig(fwhm_input, Cnt),
            mode='mirror'
        )
    else:
        mu_sctonly = muo

    if Cnt['SPN']==1:
        snno = Cnt['NSN1']
        snno_= Cnt['NSN64']
        ssrlut = axLUT['sn1_ssrb']
        saxnrm = nrmcmp['sax_f1']
    elif Cnt['SPN']==11:
        snno = Cnt['NSN11']
        snno_= snno
        ssrlut = axLUT['sn11_ssrb']
        saxnrm = nrmcmp['sax_f11']

    #LUTs for scatter
    sctLUT = get_sctLUT(scanner_params)


    #> smooth before scaling/down-sampling the mu-map and emission images
    if fwhm_input>0.:
        muim = ndi.filters.gaussian_filter(muo+muh, fwhm2sig(fwhm_input, Cnt), mode='mirror')
        emim = ndi.filters.gaussian_filter(em, fwhm2sig(fwhm_input, Cnt), mode='mirror')
    else:
        muim = muo+muh
        emim = em

    muim = ndi.interpolation.zoom( muim, Cnt['SCTSCLMU'], order=3 ) #(0.499, 0.5, 0.5)
    emim = ndi.interpolation.zoom( emim, Cnt['SCTSCLEM'], order=3 ) #(0.34, 0.33, 0.33)

    #-smooth the mu-map for mask creation.  the mask contains voxels for which attenuation ray LUT is found.
    if fwhm_input>0.:
        smomu = ndi.filters.gaussian_filter(muim, fwhm2sig(fwhm_input, Cnt), mode='mirror')
        mumsk = np.int8(smomu>0.003)
    else:
        mumsk = np.int8(muim>0.001)

    #CORE SCATTER ESTIMATION
    NSCRS, NSRNG = sctLUT['NSCRS'], sctLUT['NSRNG']
    sctout ={
        'sct_3d'  :np.zeros((Cnt['TOFBINN'], snno_, NSCRS, NSCRS), dtype=np.float32),
        'sct_val' :np.zeros((Cnt['TOFBINN'], NSRNG, NSCRS, NSRNG, NSCRS), dtype=np.float32),
    }

    #<<+>><<+>><<+>><<+>><<+>><<+>><<+>><<+>><<+>><<+>><<+>><<+>><<+>><<+>><<+>><<+>><<+>><<+>><<+>>
    nifty_scatter.vsm(sctout, muim, mumsk, emim, sctLUT, axLUT, Cnt)
    #<<+>><<+>><<+>><<+>><<+>><<+>><<+>><<+>><<+>><<+>><<+>><<+>><<+>><<+>><<+>><<+>><<+>><<+>><<+>>

    sct3d  = sctout['sct_3d']
    sctind = sctLUT['sct2aw']

    log.debug('total scatter sum: {}'.format(np.sum(sct3d)))

    #-------------------------------------------------------------------
    #> initialise output dictionary
    out = {}

    if return_uninterp:
        out['uninterp'] = sct3d
        out['indexes'] = sctind
    #-------------------------------------------------------------------


    if np.sum(sct3d)<1e-04:
        log.warning('total scatter below threshold: {}'.format(np.sum(sct3d)))
        sss    = np.zeros((snno, Cnt['NSANGLES'], Cnt['NSBINS']), dtype=np.float32)
        asnmsk = np.zeros((snno, Cnt['NSANGLES'], Cnt['NSBINS']), dtype=np.float32)
        sssr   = np.zeros((Cnt['NSEG0'], Cnt['NSANGLES'], Cnt['NSBINS']), dtype=np.float32)
        return sss, sssr, asnmsk


    # import pdb; pdb.set_trace()

    #-------------------------------------------------------------------
    if interpolate:
        #> interpolate basic scatter distributions into full size and
        #> transfer them to sinograms

        log.debug('transaxial scatter interpolation...')
        start = time.time()
        ssn, sssr = intrp_bsct(sct3d, Cnt, sctLUT, ssrlut)
        stop = time.time()
        log.debug('scatter interpolation done in {} sec.'.format(stop-start))

        if not scaling:
            out['ssrb'] = sssr
            out['sino'] = ssn
            return out
    else:
        return out
    #-------------------------------------------------------------------


    #-------------------------------------------------------------------
    # import pdb; pdb.set_trace()

    '''
    debugging scatter:
    import matplotlib.pyplot as plt
    ss = np.squeeze(sct3d)
    ss = np.sum(ss, axis=0)
    plt.matshow(ss)
    plt.matshow(sct3d[0,41,...])
    plt.matshow(np.sum(sct3d[0,0:72,...],axis=0))

    plt.plot(np.sum(sct3d, axis=(0,2,3)))

    rslt = sctout['sct_val']
    rslt.shape
    plt.matshow(rslt[0,4,:,4,:])

    debugging scatter:
    plt.matshow(np.sum(sssr, axis=(0,1)))
    plt.matshow(np.sum(ssn, axis=(0,1)))
    plt.matshow(sssr[0,70,...])
    plt.matshow(sssr[0,50,...])
    '''
    #-------------------------------------------------------------------


    #> get SSR for randoms from span-1 or span-11
    rssr = np.zeros((Cnt['NSEG0'], Cnt['NSANGLES'], Cnt['NSBINS']), dtype=np.float32)
    if scaling:
        for i in range(snno):
            rssr[ssrlut[i],:,:] += rsino[i,:,:]

    #ATTENUATION FRACTIONS for scatter only regions, and NORMALISATION for all SCATTER
    #<<+>><<+>><<+>><<+>><<+>><<+>><<+>><<+>><<+>><<+>><<+>><<+>><<+>><<+>><<+>><<+>><<+>><<+>>
    currentspan = Cnt['SPN']
    Cnt['SPN'] = 1
    atto = np.zeros((txLUT['Naw'], Cnt['NSN1']), dtype=np.float32)
    petprj.fprj(atto, mu_sctonly, txLUT, axLUT, np.array([-1], dtype=np.int32), Cnt, 1)
    atto = mmraux.putgaps(atto, txLUT, Cnt)
    #--------------------------------------------------------------
    #> get norm components setting the geometry and axial to ones as they are accounted for differently
    nrmcmp['geo'][:] = 1
    nrmcmp['axe1'][:] = 1
    #get sino with no gaps
    nrmg = np.zeros((txLUT['Naw'], Cnt['NSN1']), dtype=np.float32)
    mmr_auxe.norm(nrmg, nrmcmp, histo['buckets'], axLUT, txLUT['aw2ali'], Cnt)
    nrm = mmraux.putgaps(nrmg, txLUT, Cnt)
    #--------------------------------------------------------------


    #> get attenuation + norm in (span-11) and SSR
    attossr = np.zeros((Cnt['NSEG0'], Cnt['NSANGLES'], Cnt['NSBINS']), dtype=np.float32)
    nrmsssr = np.zeros((Cnt['NSEG0'], Cnt['NSANGLES'], Cnt['NSBINS']), dtype=np.float32)

    for i in range(Cnt['NSN1']):
        si = axLUT['sn1_ssrb'][i]
        attossr[si,:,:] += atto[i,:,:] / float(axLUT['sn1_ssrno'][si])
        nrmsssr[si,:,:] += nrm[i,:,:] / float(axLUT['sn1_ssrno'][si])
    if currentspan==11:
        Cnt['SPN']=11
        nrmg = np.zeros((txLUT['Naw'], snno), dtype=np.float32)
        mmr_auxe.norm(nrmg, nrmcmp, histo['buckets'], axLUT, txLUT['aw2ali'], Cnt)
        nrm = mmraux.putgaps(nrmg, txLUT, Cnt)
    #--------------------------------------------------------------

    #<<+>><<+>><<+>><<+>><<+>><<+>><<+>><<+>><<+>><<+>><<+>><<+>><<+>><<+>><<+>><<+>><<+>><<+>>

    #get the mask for the object from uncorrected emission image
    if emmsk and os.path.isfile(datain['em_nocrr']):
        nim = nib.load(datain['em_nocrr'])
        A   = nim.get_sform()
        eim = nim.get_fdata(dtype=np.float32)
        eim = eim[:,::-1,::-1]
        eim = np.transpose(eim, (2, 1, 0))

        em_sctonly = ndi.filters.gaussian_filter(eim, fwhm2sig(.6, Cnt), mode='mirror')
        msk = np.float32(em_sctonly>0.07*np.max(em_sctonly))
        msk = ndi.filters.gaussian_filter(msk, fwhm2sig(.6, Cnt), mode='mirror')
        msk = np.float32(msk>0.01)
        msksn = mmrprj.frwd_prj(msk, txLUT, axLUT, Cnt)

        mssr = mmraux.sino2ssr(msksn, axLUT, Cnt)
        mssr = mssr>0
    else:
        mssr = np.zeros((Cnt['NSEG0'], Cnt['NSANGLES'], Cnt['NSBINS']), dtype=np.bool);

    #<<+>><<+>><<+>><<+>><<+>><<+>><<+>><<+>><<+>><<+>><<+>><<+>><<+>><<+>><<+>><<+>><<+>><<+>>


    #======== SCALING ========
    #> scale scatter using non-TOF SSRB sinograms

    #> gap mask
    rmsk = (txLUT['msino']>0).T
    rmsk.shape = (1,Cnt['NSANGLES'],Cnt['NSBINS'])
    rmsk = np.repeat(rmsk, Cnt['NSEG0'], axis=0)

    #> include attenuating object into the mask (and the emission if selected)
    amsksn = np.logical_and( attossr>=mask_threshlod, rmsk) * ~mssr

    #> scaling factors for SSRB scatter
    scl_ssr = np.zeros( (Cnt['NSEG0']), dtype=np.float32)

    for sni in range(Cnt['NSEG0']):
        #> region for scaling defined by the percentage of lowest
        #> but usable/significant scatter
        thrshld = prcnt_scl * np.max(sssr[sni,:,:])
        amsksn[sni,:,:] *= (sssr[sni,:,:]>thrshld)
        amsk = amsksn[sni,:,:]

        #> normalised estimated scatter
        mssn = sssr[sni,:,:] * nrmsssr[sni,:,:]
        vpsn = histo['pssr'][sni, amsk] - rssr[sni, amsk]
        scl_ssr[sni] = np.sum(vpsn) / np.sum(mssn[amsk])

        #> scatter SSRB sinogram output
        sssr[sni,:,:] *= nrmsssr[sni,:,:]*scl_ssr[sni]


    #=== scale scatter for the full-size sinogram ===
    sss = np.zeros((snno, Cnt['NSANGLES'], Cnt['NSBINS']), dtype=np.float32);
    for i in range(snno):
        sss[i,:,:] = ssn[i,:,:]*scl_ssr[ssrlut[i]]*saxnrm[i] * nrm[i,:,:]


    '''
    #> debug
    si = 60
    ai = 60
    matshow(sssr[si,...])

    figure()
    plot(histo['pssr'][si,ai,:])
    plot(rssr[si,ai,:]+sssr[si,ai,:])

    plot(np.sum(histo['pssr'],axis=(0,1)))
    plot(np.sum(rssr+sssr,axis=(0,1)))
    '''


    #=== OUTPUT ===
    if return_uninterp:
        out['uninterp'] = sct3d
        out['indexes'] = sctind

    if return_ssrb:
        out['ssrb'] = sssr
        out['rssr'] = rssr

    if return_mask:
        out['mask'] = amsksn

    if return_scaling:
        out['scaling'] = scl_ssr

    # if self_scaling:
    #     out['scl_sn1'] = scl_ssn


    if not out:
        return sss
    else:
        out['sino'] = sss
        return out



"""module for pipelined image reconstruction and analysis"""
__author__      = ("Pawel J. Markiewicz", "Casper O. da Costa-Luis")
__copyright__   = "Copyright 2020"

import logging
import os
import sys
from numbers import Integral
from subprocess import call
from textwrap import dedent

import numpy as np
import scipy.ndimage as ndi

from niftypet import nimpa

from ..lm import dynamic_timings
from ..lm.mmrhist import mmrhist
from ..prj import mmrrec
from . import obtain_image
from .mmrimg import image_affine

log = logging.getLogger(__name__)


def mmrchain(
    datain,         # all input data in a dictionary
    scanner_params, # all scanner parameters in one dictionary
                    # containing constants, transaxial and axial
                    # LUTs.
    outpath='',     # output path for results
    frames=['fluid', [0,0]], # definition of time frames.
    mu_h = [],      # hardware mu-map.
    mu_o = [],      # object mu-map.
    tAffine = None, # affine transformations for the mu-map for
                    # each time frame separately.

    itr=4,          # number of OSEM iterations
    fwhm=0.,        # Gaussian Post-Smoothing FWHM
    psf=None,       # Resolution Modelling
    recmod = -1,    # reconstruction mode: -1: undefined, chosen
                    # automatically. 3: attenuation and scatter
                    # correction, 1: attenuation correction
                    # only, 0: no correction (randoms only).
    histo=None,     # input histogram (from list-mode data);
                    # if not given, it will be performed.

    decay_ref_time=None, # decay corrects relative to the reference
                    # time provided; otherwise corrects to the scan
                    # start time.

    trim=False,
    trim_scale=2,
    trim_interp=0,  # interpolation for upsampling used in PVC
    trim_memlim=True,   # reduced use of memory for machines
                        # with limited memory (slow though)

    pvcroi=[],      # ROI used for PVC.  If undefined no PVC
                    # is performed.

    pvcreg_tool = 'niftyreg', # the registration tool used in PVC
    store_rois = False, # stores the image of PVC ROIs
                        # as defined in pvcroi.

    pvcpsf=[],
    pvcitr=5,

    fcomment='',    # text comment used in the file name of
                    # generated image files
    ret_sinos=False,# return prompt, scatter and randoms
                    # sinograms for each reconstruction
    ret_histo=False,# return histogram (LM processing output) for
                    # each image frame
    store_img = True,
    store_img_intrmd=False,
    store_itr=[],   # store any reconstruction iteration in
                    # the list.  ignored if the list is empty.
    del_img_intrmd=False,
):
    # decompose all the scanner parameters and constants
    Cnt   = scanner_params['Cnt']
    txLUT = scanner_params['txLUT']
    axLUT = scanner_params['axLUT']

    # -------------------------------------------------------------------------
    # HISOTGRAM PRECEEDS FRAMES
    if not histo==None and 'psino' in histo:
        frames = ['fluid', [histo['t0'], histo['t1']]]
    else:
        histo = None
        log.warning('the given histogram does not contain a prompt sinogram--will generate a histogram.')

    # FRAMES
    # check for the provided dynamic frames
    if isinstance(frames, list):
        # Can be given in three ways:
        # * a 1D list (duration of each frame is listed)
        # * a more concise 2D list--repetition and duration lists in
        #   each entry.  Must start with the 'def' entry.
        # * a 2D list with fluid timings: must start with the string
        #   'fluid' or 'timings'.  a 2D list with consecutive lists
        #   describing start and end of the time frame, [t0, t1];
        #   The number of time frames for this option is unlimited,
        #   provided the t0 and t1 are within the acquisition times.

        # 2D starting with entry 'fluid' or 'timings'
        if (isinstance(frames[0], str) and frames[0] in ('fluid', 'timings')
            and all([isinstance(t, list) and len(t) == 2 for t in frames[1:]])):
            t_frms = frames[1:]
        # if 2D definitions, starting with entry 'def':
        elif (isinstance(frames[0], str) and frames[0]=='def'
              and all([isinstance(t,list) and len(t)==2 for t in frames[1:]])):
            # get total time and list of all time frames
            dfrms = dynamic_timings(frames)
            t_frms = dfrms[1:]

        # if 1D:
        elif all([isinstance(t, Integral) for t in frames]):
            # get total time and list of all time frames
            dfrms = dynamic_timings(frames)
            t_frms = dfrms[1:]

        else:
            log.error('osemdyn: frames definitions are not given\
                in the correct list format: 1D [15,15,30,30,...]\
                or 2D list [[2,15], [2,30], ...]')
    else:
        log.error('provided dynamic frames definitions are incorrect (should be a list of definitions).')
        raise TypeError('Wrong data type for dynamic frames')
    # number of dynamic time frames
    nfrm = len(t_frms)
    # -------------------------------------------------------------------------



    # -------------------------------------------------------------------------
    # create folders for results
    if outpath=='':
        petdir = os.path.join(datain['corepath'], 'reconstructed')
        fmudir = os.path.join(datain['corepath'], 'mumap-obj')
        pvcdir = os.path.join(datain['corepath'], 'PRCL')
    else:
        petdir = os.path.join(outpath, 'PET')
        fmudir = os.path.join(outpath, 'mumap-obj')
        pvcdir = os.path.join(outpath, 'PRCL')

    # folder for co-registered mu-maps (for motion compensation)
    fmureg = os.path.join( fmudir, 'registered')
    # folder for affine transformation MR/CT->PET
    petaff = os.path.join( petdir, 'faffine')

    # folder for reconstructed images (dynamic or static depending on number of frames).
    if nfrm>1:
        petimg = os.path.join(petdir, 'multiple-frames')
        pvcdir = os.path.join(pvcdir, 'multiple-frames')
    elif nfrm==1:
        petimg = os.path.join(petdir, 'single-frame')
        pvcdir = os.path.join(pvcdir, 'single-frame')
    else:
        raise TypeError('Unrecognised/confusing time frames!')
    # create now the folder
    nimpa.create_dir(petimg)
    # create folder
    nimpa.create_dir(petdir)
    # -------------------------------------------------------------------------


    # -------------------------------------------------------------------------
    # MU-MAPS
    # get the mu-maps, if given;  otherwise will use blank mu-maps.
    if not tAffine is None:
        muod = obtain_image(mu_o, imtype='object mu-map')
    else:
        muod = obtain_image(mu_o, Cnt=Cnt, imtype='object mu-map')

    # hardware mu-map
    muhd = obtain_image(mu_h, Cnt, imtype='hardware mu-map')

    # choose the mode of reconstruction based on the provided (or not) mu-maps
    if muod['exists'] and muhd['exists'] and recmod==-1:
        recmod = 3
    elif  (muod['exists'] or muhd['exists']) and recmod==-1:
        recmod = 1
        log.warning('partial mu-map:  scatter correction is switched off.')
    else:
        if recmod==-1:
            recmod = 0
            log.warning('no mu-map provided: scatter and attenuation corrections are switched off.')
    # -------------------------------------------------------------------------

    #import pdb; pdb.set_trace()

    # output dictionary
    output = {}
    output['recmod'] = recmod
    output['frames'] = t_frms
    output['#frames'] = nfrm

    # if affine transformation is given the baseline mu-map in NIfTI file or dictionary has to be given
    if tAffine is None:
        log.info('using the provided mu-map the same way for all frames.')
    else:
        if len(tAffine)!=nfrm:
            log.error('the number of affine transformations in the list\
                has to be the same as the number of dynamic frames!')
            raise ValueError('Inconsistent number of frames.')
        elif not isinstance(tAffine, list):
            log.error('tAffine has to be a list of either 4x4 numpy arrays\
                of affine transformations or a list of file path strings!')
            raise ValueError('Expecting a list.')
        elif 'fim' not in muod:
            log.error('when tAffine is given, the object mu-map has to be\
                provided either as a dictionary or NIfTI file!')
            raise NameError('No path to object mu-map.')

        # check if all are file path strings to the existing files
        if all([isinstance(t, str) for t in tAffine]):
            if all([os.path.isfile(t) for t in tAffine]):
                # the internal list of affine transformations
                faff_frms = tAffine
                log.info('using provided paths to affine transformations for each dynamic frame.')
            else:
                log.error('not all provided paths are valid!')
                raise IOError('Wrong paths.')
        # check if all are numpy arrays
        elif all([isinstance(t, (np.ndarray, np.generic)) for t in tAffine]):
            # create the folder for dynamic affine transformations
            nimpa.create_dir(petaff)
            faff_frms = []
            for i in range(nfrm):
                fout = os.path.join(petaff, 'affine_frame('+str(i)+').txt')
                np.savetxt(fout, tAffine[i], fmt='%3.9f')
                faff_frms.append(fout)
            log.info('using provided numpy arrays affine transformations for each dynamic frame.')
        else:
            raise ValueError('Affine transformations for each dynamic frame could not be established.')

        # -------------------------------------------------------------------------------------
        # get ref image for mu-map resampling
        # -------------------------------------------------------------------------------------
        if 'fmuref' in muod:
            fmuref = muod['fmuref']
            log.info('reusing the reference mu-map from the object mu-map dictionary.')
        else:
            # create folder if doesn't exists
            nimpa.create_dir(fmudir)
            # ref file name
            fmuref = os.path.join(fmudir, 'muref.nii.gz')
            # ref affine
            B = image_affine(datain, Cnt, gantry_offset=False)
            # ref image (blank)
            im = np.zeros((Cnt['SO_IMZ'], Cnt['SO_IMY'], Cnt['SO_IMX']), dtype=np.float32)
            # store ref image
            nimpa.array2nii(im, B, fmuref)
            log.info('generated a reference mu-map in:\n{}'.format(fmuref))
        # -------------------------------------------------------------------------------------

        output['fmuref'] = fmuref
        output['faffine'] = faff_frms

    # output list of intermediate file names for mu-maps and PET images (useful for dynamic imaging)
    if not tAffine is None: output['fmureg'] = []

    if store_img_intrmd:
        output['fpeti'] = []
        if fwhm>0:
            output['fsmoi'] = []

    # dynamic images in one numpy array
    dynim = np.zeros((nfrm, Cnt['SO_IMZ'], Cnt['SO_IMY'], Cnt['SO_IMY']), dtype=np.float32)
    #if asked, output only scatter+randoms sinogram for each frame
    if ret_sinos and itr>1 and recmod>2:
        dynmsk = np.zeros((nfrm, Cnt['NSEG0'], Cnt['NSANGLES'], Cnt['NSBINS']), dtype=np.float32)
        dynrsn = np.zeros((nfrm, Cnt['NSN11'], Cnt['NSANGLES'], Cnt['NSBINS']), dtype=np.float32)
        dynssn = np.zeros((nfrm, Cnt['NSN11'], Cnt['NSANGLES'], Cnt['NSBINS']), dtype=np.float32)
        dynpsn = np.zeros((nfrm, Cnt['NSN11'], Cnt['NSANGLES'], Cnt['NSBINS']), dtype=np.float32)

    #> returning dictionary of histograms if requested
    if ret_histo:
        hsts = {}


    # import pdb; pdb.set_trace()

    # starting frame index with reasonable prompt data
    ifrmP = 0
    # iterate over frame index
    for ifrm in range(nfrm):
        # start time of a current (ifrm-th) dynamic frame
        t0 = int(t_frms[ifrm][0])
        # end time of a current (ifrm-th) dynamic frame
        t1 = int(t_frms[ifrm][1])
        # --------------
        # check if there is enough prompt data to do a reconstruction
        # --------------
        log.info('dynamic frame times t0={}, t1={}:'.format(t0, t1))
        if histo==None:
            hst = mmrhist(datain, scanner_params, t0=t0, t1=t1)
        else:
            hst = histo
            log.info(dedent('''\
                ------------------------------------------------------
                using provided histogram
                ------------------------------------------------------'''))

        if ret_histo:
            hsts[str(t0)+'-'+str(t1)] = hst

        if np.sum(hst['dhc'])>0.99*np.sum(hst['phc']):
            log.warning(dedent('''\
                ===========================================================================
                amount of randoms is the greater part of prompts => omitting reconstruction
                ==========================================================================='''))
            ifrmP = ifrm+1
            continue
        # --------------------
        # transform the mu-map if given the affine transformation for each frame
        if not tAffine is None:
            # create the folder for aligned (registered for motion compensation) mu-maps
            nimpa.create_dir(fmureg)
            # the converted nii image resample to the reference size
            fmu = os.path.join(fmureg, 'mumap_dyn_frm'+str(ifrm)+fcomment+'.nii.gz')
            # command for resampling
            if os.path.isfile( Cnt['RESPATH'] ):
                cmd = [Cnt['RESPATH'],
                '-ref', fmuref,
                '-flo', muod['fim'],
                '-trans', faff_frms[ifrm],
                '-res', fmu,
                '-pad', '0']
                if log.getEffectiveLevel() > log.INFO:
                    cmd.append('-voff')
                call(cmd)
            else:
                raise IOError('Incorrect path to NiftyReg (resampling) executable.')
            # get the new mu-map from the just resampled file
            muodct = nimpa.getnii(fmu, output='all')
            muo = muodct['im']
            A = muodct['affine']
            muo[muo<0] = 0
            output['fmureg'].append(fmu)
        else:
            muo = muod['im']
        #---------------------

        # output image file name
        if nfrm>1:
            frmno = '_frm'+str(ifrm)
        else:
            frmno = ''

        # run OSEM reconstruction of a single time frame
        recimg = mmrrec.osemone(datain, [muhd['im'], muo],
                                hst, scanner_params,
                                decay_ref_time=decay_ref_time,
                                recmod=recmod, itr=itr, fwhm=fwhm, psf=psf,
                                outpath=petimg,
                                frmno=frmno,
                                fcomment=fcomment+'_i',
                                store_img=store_img_intrmd,
                                store_itr=store_itr,
                                ret_sinos=ret_sinos)

        # form dynamic Numpy array
        if fwhm>0:
            dynim[ifrm,:,:,:] = recimg.imsmo
        else:
            dynim[ifrm,:,:,:] = recimg.im

        if ret_sinos and itr>1 and recmod>2:
            dynpsn[ifrm,:,:,:] = np.squeeze(hst['psino'])
            dynssn[ifrm,:,:,:] = np.squeeze(recimg.ssn)
            dynrsn[ifrm,:,:,:] = np.squeeze(recimg.rsn)
            dynmsk[ifrm,:,:,:] = np.squeeze(recimg.amsk)

        if store_img_intrmd:
            output['fpeti'].append(recimg.fpet)
            if fwhm>0:
                output['fsmoi'].append(recimg.fsmo)

        if nfrm==1: output['tuple'] = recimg

    output['im'] = np.squeeze(dynim)

    if ret_sinos and itr>1 and recmod>2:
        output['sinos'] = {
            'psino':dynpsn,
            'ssino':dynssn,
            'rsino':dynrsn,
            'amask':dynmsk}

    if ret_histo:
        output['hst'] = hsts


    # ----------------------------------------------------------------------
    # trim the PET image
    # images have to be stored for PVC
    if pvcroi: store_img_intrmd = True
    if trim:
        # create file name
        if 'lm_dcm' in datain:
            fnm = os.path.basename(datain['lm_dcm'])[:20]
        elif 'lm_ima' in datain:
            fnm = os.path.basename(datain['lm_ima'])[:20]
        # trim PET and upsample
        petu = nimpa.imtrimup(
            dynim,
            affine=image_affine(datain, Cnt),
            scale=trim_scale,
            int_order=trim_interp,
            outpath=petimg,
            fname = fnm,
            fcomment=fcomment,
            store_img_intrmd=store_img_intrmd,
            memlim=trim_memlim,
            verbose=log.getEffectiveLevel()
        )

        output.update({'trimmed': { 'im':petu['im'],
                                    'fpet':petu['fimi'],
                                    'affine':petu['affine']}}
        )
    # ----------------------------------------------------------------------


    # ----------------------------------------------------------------------
    #run PVC if requested and required input given
    if pvcroi:
        if not os.path.isfile(datain['T1lbl']):
            raise Exception('No labels and/or ROIs image definitions found!')
        else:
            # get the PSF kernel for PVC
            if not pvcpsf:
                pvcpsf = nimpa.psf_measured(scanner='mmr', scale=trim_scale)
            else:
                if isinstance(pvcpsf, (np.ndarray, np.generic)) and pvcpsf.shape!=(3, 2*Cnt['RSZ_PSF_KRNL']+1):
                    raise ValueError('the PSF kernel has to be an numpy array with the shape of ({},{})'.format(3, 2*Cnt['RSZ_PSF_KRNL']+1))

        #> file names for NIfTI images of PVC ROIs and PVC corrected PET
        froi = []
        fpvc = []

        #> perform PVC for each time frame
        dynpvc = np.zeros(petu['im'].shape, dtype=np.float32)
        for i in range(ifrmP,nfrm):
            # transform the parcellations (ROIs) if given the affine transformation for each frame
            if tAffine is None:
                log.warning('affine transformation are not provided: will generate for the time frame.')
                faffpvc = None
                #raise StandardError('No affine transformation')
            else:
                faffpvc = faff_frms[i]

            # chose file name of individual PVC images
            if nfrm>1:
                fcomment_pvc = '_frm'+str(i)+fcomment
            else:
                fcomment_pvc = fcomment
            #============================
            # perform PVC
            petpvc_dic = nimpa.pvc_iyang(
                petu['fimi'][i],
                datain,
                Cnt,
                pvcroi,
                pvcpsf,
                tool=pvcreg_tool,
                itr=pvcitr,
                faff=faffpvc,
                fcomment=fcomment_pvc,
                outpath=pvcdir,
                store_rois=store_rois,
                store_img=store_img_intrmd)
            #============================
            if nfrm>1:
                dynpvc[i,:,:,:] = petpvc_dic['im']
            else:
                dynpvc = petpvc_dic['im']
            fpvc.append(petpvc_dic['fpet'])

            if store_rois: froi.append(petpvc_dic['froi'])

        #> update output dictionary
        output.update({'impvc':dynpvc})
        output['fprc'] = petpvc_dic['fprc']
        output['imprc'] = petpvc_dic['imprc']

        if store_img_intrmd: output.update({'fpvc':fpvc})
        if store_rois: output.update({'froi':froi})
    # ----------------------------------------------------------------------

    if store_img:
        # description for saving NIFTI image
        # attenuation number: if only bed present then it is 0.5
        attnum =  ( 1*muhd['exists'] + 1*muod['exists'] ) / 2.
        descrip =    'alg=osem'                     \
                    +';att='+str(attnum*(recmod>0)) \
                    +';sct='+str(1*(recmod>1))      \
                    +';spn='+str(Cnt['SPN'])        \
                    +';sub=14'                      \
                    +';itr='+str(itr)               \
                    +';fwhm='+str(fwhm)             \
                    +';psf='+str(psf)       \
                    +';nfrm='+str(nfrm)

        # squeeze the not needed dimensions
        dynim = np.squeeze(dynim)

        # NIfTI file name for the full PET image (single or multiple frame)

        # save the image to NIfTI file
        if nfrm==1:
            t0 = hst['t0']
            t1 = hst['t1']
            if t1==t0:
                t0 = 0
                t1 = hst['dur']
            fpet = os.path.join(
                    petimg,
                    os.path.basename(recimg.fpet)[:8] \
                    +'_t-'+str(t0)+'-'+str(t1)+'sec' \
                    +'_itr-'+str(itr) )
            fpeto = fpet+fcomment+'.nii.gz'
            nimpa.prc.array2nii( dynim[::-1,::-1,:], recimg.affine, fpeto, descrip=descrip)
        else:
            fpet = os.path.join(
                    petimg,
                    os.path.basename(recimg.fpet)[:8]\
                    +'_nfrm-'+str(nfrm)+'_itr-'+str(itr)
                )
            fpeto = fpet+fcomment+'.nii.gz'
            nimpa.prc.array2nii( dynim[:,::-1,::-1,:], recimg.affine, fpeto, descrip=descrip)

        # get output file names for trimmed/PVC images
        if trim:
            # folder for trimmed and dynamic
            pettrim = os.path.join( petimg, 'trimmed')
            # make folder
            nimpa.create_dir(pettrim)
            # trimming scale added to NIfTI descritoption
            descrip_trim = descrip + ';trim_scale='+str(trim_scale)
            # file name for saving the trimmed image
            fpetu = os.path.join(pettrim, os.path.basename(fpet) + '_trimmed-upsampled-scale-'+str(trim_scale))
            # in case of PVC
            if pvcroi:
                # itertive Yang (iY) added to NIfTI descritoption
                descrip_pvc = descrip_trim + ';pvc=iY'
                # file name for saving the PVC NIfTI image
                fpvc = fpetu + '_PVC' + fcomment + '.nii.gz'
                output['trimmed']['fpvc'] = fpvc

            # update the trimmed image file name
            fpetu += fcomment+'.nii.gz'
            # store the file name in the output dictionary
            output['trimmed']['fpet'] = fpetu

        output['fpet'] = fpeto

        # save images
        if nfrm==1:
            if trim:
                nimpa.prc.array2nii( petu['im'][::-1,::-1,:], petu['affine'], fpetu, descrip=descrip_trim)
            if pvcroi:
                nimpa.prc.array2nii( dynpvc[::-1,::-1,:], petu['affine'], fpvc, descrip=descrip_pvc)
        elif nfrm>1:
            if trim:
                nimpa.prc.array2nii( petu['im'][:,::-1,::-1,:], petu['affine'], fpetu, descrip=descrip_trim)
            if pvcroi:
                nimpa.prc.array2nii( dynpvc[:,::-1,::-1,:], petu['affine'], fpvc, descrip=descrip_pvc)


    if del_img_intrmd:
        if pvcroi:
            for fi in fpvc:
                os.remove(fi)
        if trim:
            for fi in petu['fimi']:
                os.remove(fi)




    return output



import logging

from niftypet.ninst import install_tools as tls

from .conftest import HOME

log = logging.getLogger(__name__)
DATA_URL = "https://zenodo.org/record/3877529/files/amyloidPET_FBP_TP0_extra.zip?download=1"


def main():
    log.info(f"Downloading {DATA_URL}\nto ${{DATA_ROOT:-~}} ({HOME})")
    with tls.urlopen_cached(DATA_URL, HOME) as fd:
        tls.extractall(fd, HOME)


if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO)
    main()
